# reference：https://github.com/maxtheman/opengpts/blob/d3425b1ba80aec48953a327ecd9a61b80efb0e69/backend/app/agent_types/openai_agent.py
import json

from langchain.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool
from langchain_core.language_models.base import LanguageModelLike
from langchain_core.messages import  SystemMessage, ToolMessage
from langgraph.graph import END
from langgraph.graph.message import MessageGraph
from langgraph.prebuilt import ToolExecutor, ToolInvocation
from typing import Any


def create_openai_func_call_agent_executor(tools: list[BaseTool], llm: LanguageModelLike,
                                        system_message: str, **kwargs):

    async def _get_messages(messages):
        msgs = []
        for m in messages:
            if isinstance(m, ToolMessage):
                _dict = m.dict()
                _dict['content'] = str(_dict['content'])
                m_c = ToolMessage(**_dict)
                msgs.append(m_c)
            else:
                msgs.append(m)

        return [SystemMessage(content=system_message)] + msgs

    if tools:
        llm_with_tools = llm.bind(tools=[convert_to_openai_tool(t) for t in tools])
    else:
        llm_with_tools = llm
    agent = _get_messages | llm_with_tools
    tool_executor = ToolExecutor(tools)

    # Define the function that determines whether to continue or not
    def should_continue(messages):
        # If there is no FuncCall, then we finish
        last_message = messages[-1]
        if not last_message.tool_calls:
            return "end"
        # Otherwise if there is, we continue
        else:
            return "continue"

    # Define the function to execute tools
    async def call_tool(messages):
        actions: list[ToolInvocation] = []
        # Based on the continue condition
        # we know the last message involves a FuncCall
        last_message = messages[-1]
        for tool_call in last_message.additional_kwargs['tool_calls']:
            function = tool_call['function']
            function_name = function['name']
            _tool_input = json.loads(function['arguments'] or '{}')
            # We construct an ToolInvocation from the function_call
            actions.append(ToolInvocation(
                tool=function_name,
                tool_input=_tool_input,
            ))
        # We call the tool_executor and get back a response
        responses = await tool_executor.abatch(actions, **kwargs)
        # We use the response to create a ToolMessage
        tool_messages = [
            ToolMessage(
                tool_call_id=tool_call['id'],
                content=response,
                additional_kwargs={'name': tool_call['function']['name']},
            )
            for tool_call, response in zip(last_message.additional_kwargs['tool_calls'], responses)
        ]
        return tool_messages

    workflow = MessageGraph()

    # Define the two nodes we will cycle between
    workflow.add_node('agent', agent)
    workflow.add_node('action', call_tool)

    # Set the entrypoint as `agent`
    # This means that this node is the first one called
    workflow.set_entry_point('agent')

    # We now add a conditional edge
    workflow.add_conditional_edges(
        # First, we define the start node. We use `agent`.
        # This means these are the edges taken after the `agent` node is called.
        'agent',
        # Next, we pass in the function that will determine which node is called next.
        should_continue,
        # Finally we pass in a mapping.
        # The keys are strings, and the values are other nodes.
        # END is a special node marking that the graph should finish.
        # What will happen is we will call `should_continue`, and then the output of that
        # will be matched against the keys in this mapping.
        # Based on which one it matches, that node will then be called.
        {
            # If `tools`, then we call the tool node.
            'continue': 'action',
            # Otherwise we finish.
            'end': END,
        },
    )

    # We now add a normal edge from `tools` to `agent`.
    # This means that after `tools` is called, `agent` node is called next.
    workflow.add_edge('action', 'agent')

    # Finally, we compile it!
    # This compiles it into a LangChain Runnable,
    # meaning you can use it as you would any other runnable
    return workflow.compile()



import argparse
import sys
import json

def read_arg(param, is_path=False):
    args = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument(param, nargs='?', help=f'Parameter {param}')
    args = parser.parse_known_args(args)[0]
    value = getattr(args, param.lstrip('-'))
    # 如果是路径并且有引号，去掉引号
    if is_path and value and value.startswith('"') and value.endswith('"'):
        value = value[1:-1]

    return value

def read_params():
  params_file = read_arg('params_file', True)
  params = None
  try:
      with open(params_file, 'r', encoding='utf-8') as f:
          params = json.load(f)
  except Exception as e:
      print(e)
  return params



import sys
import os
import argparse
import importlib
from langchain.agents import tool as _tool
import types
import langchain_core

def wrapped_tool(tool, module_path):
    old_tool_fun = tool.func
    def _wrapped_tool(*args, **kwargs):
        print(f"调用了{tool.name}")
        old_path = os.getcwd()
        try:
            sys.path.insert(0, module_path)  # 插入到第一个位置
            os.chdir(module_path)
            result = old_tool_fun(*args, **kwargs)
        finally:
            sys.path.remove(module_path)
            os.chdir(old_path)
        print(f"{tool.name}调用完毕,结果为:", result)
        return result
    return _wrapped_tool


def import_tools(tool_id):
    tool_path = "../../tool"
    old_path = os.getcwd()
    module_path = os.path.join(tool_path, tool_id)
    module_path = os.path.normpath(os.path.abspath(module_path))

    if not os.path.exists(module_path):
        print(f"Tool {tool_id} not found")
        return []

    # 保存当前的 sys.modules 状态
    original_modules = sys.modules.copy()

    try:
        sys.path.insert(0, module_path)  # 插入到第一个位置
        os.chdir(module_path)
        module = importlib.import_module("tool")
        export_tools = getattr(module, "export_tools")
        temp_list = []
        for tool in export_tools:
            assert isinstance(tool, langchain_core.tools.StructuredTool) or isinstance(tool, types.FunctionType), f"Tool is not a StructuredTool or function: {tool}"
            if isinstance(tool, types.FunctionType) and not isinstance(tool, langchain_core.tools.StructuredTool):
                temp_list.append(_tool(tool))
        export_tools = temp_list

    except Exception as e:
        print(f"Error loading tool {tool_id}: {e}")
        sys.path.remove(module_path)
        os.chdir(old_path)
        return []

    # 卸载模块并恢复 sys.modules 状态
    importlib.invalidate_caches()
    for key in list(sys.modules.keys()):
        if key not in original_modules:
            del sys.modules[key]

    sys.path.remove(module_path)
    os.chdir(old_path)

    for tool in export_tools:
        tool.func = wrapped_tool(tool, module_path)

    return export_tools


def read_arg(param: str, is_path=False):
    parser = argparse.ArgumentParser()
    parser.add_argument(param, nargs='?', help=f'Parameter {param}')
    args = parser.parse_args()
    value = getattr(args, param.lstrip('-'))
    # 如果是路径并且有引号，去掉引号
    if is_path and value and value.startswith('"') and value.endswith('"'):
        value = value[1:-1]

    return value



import asyncio
import json
from langchain_openai import ChatOpenAI
from .chat import info
from .assistant import Runnable
from langchain_core.messages import  HumanMessage, AIMessage,AIMessageChunk
from langchain_core.agents import AgentFinish



async def run_agent(agent, inputs: list):
    if agent.agent_execution_mode == 'FuncCall':
        exist_ids = set()
        content_list = []
        async for event in agent.astream_events(inputs, version="v2"):
            kind = event["event"]
            if kind == "on_chat_model_stream":
                content = event["data"]["chunk"].content
                if content:
                    content_list.append(content)
                    info("assistant", content)
            elif kind == "on_chain_start":
                data = event["data"]
                if "input" in data:
                    input_list = data["input"] if isinstance(data["input"], list) else [data["input"]]
                    msg = input_list[-1]
                    if isinstance(msg, AIMessage) and not isinstance(msg, AIMessageChunk):
                        if "tool_calls" in msg.additional_kwargs:
                            tool_calls = msg.additional_kwargs["tool_calls"]
                            for t in tool_calls:
                                if t["id"] in exist_ids:
                                    continue
                                exist_ids.add(t["id"])
                                tool = t["function"]["name"]
                                info("assistant", f"[调用工具:{tool}]")

        return "".join(content_list)
    else:
        assert agent.agent_execution_mode == 'ReAct'
        inputs_msg = {'input': inputs[-1].content,'chat_history': inputs[:-1]}
        use_tools = []
        async for event in agent.astream_events(inputs_msg, version="v2"):
            kind = event["event"]
            result = None
            if kind == "on_chain_end":
                if 'data' in event:
                    if 'output' in event['data']:
                        output = event['data']['output']
                        if 'agent_outcome' in output and "input" in output:
                            outcome = output['agent_outcome']
                            if isinstance(outcome, AgentFinish):
                                result = outcome.return_values['output']
            elif kind == "on_tool_start":
                use_tools.append(event['name'])
                info("assistant", f"[调用工具:{use_tools}]")
        return result


if __name__ == "__main__":
    from langchain_experimental.tools import PythonREPLTool
    llm_file = r"C:\Users\aiqqq\AppData\Roaming\pycoze\JsonStorage\llm.json"
    with open(llm_file, "r", encoding="utf-8") as f:
        cfg = json.load(f)
        chat = ChatOpenAI(api_key=cfg["apiKey"], base_url=cfg['baseURL'], model=cfg["model"], temperature=0)
    python_tool = PythonREPLTool()
    agent = Runnable(agent_execution_mode='FuncCall', # 'FuncCall' or 'ReAct'，大模型支持FuncCall的话就用FuncCall
                                               tools=[python_tool],
                                               llm=chat,
                                               assistant_message="请以女友的口吻回答，输出不小于100字，可以随便说点其他的",)

    inputs = [HumanMessage(content="计算根号7+根号88")]
    print(asyncio.run(run_agent(agent, inputs)))



import json
from langchain_openai import ChatOpenAI
from .base import import_tools
from .agent import run_agent, Runnable, INPUT_MESSAGE, output
import asyncio
from langchain_core.messages import HumanMessage
from pycoze import utils


params = utils.arg.read_params()
llm_file = params["appPath"] + "/JsonStorage/llm.json"

def load_role_setting(bot_setting_file:str):
    with open(bot_setting_file, "r", encoding="utf-8") as f:
        return json.load(f)

def load_tools(bot_setting_file:str):
    with open(bot_setting_file, "r", encoding="utf-8") as f:
        role_setting = json.load(f)

    tools = []
    for tool_id in role_setting["tools"]:
        tools.extend(import_tools(tool_id))
    return tools




def chat(bot_setting_file:str):
    history = []

    while True:
        message = input()
        role_setting = load_role_setting(bot_setting_file)
        tools = load_tools(bot_setting_file)
        if not message.startswith(INPUT_MESSAGE):
            raise ValueError("Invalid message")
        message = json.loads(message[len(INPUT_MESSAGE):])["content"]
        print("user:", message)

        with open(llm_file, "r", encoding="utf-8") as f:
            cfg = json.load(f)
            chat = ChatOpenAI(api_key=cfg["apiKey"], base_url=cfg['baseURL'], model=cfg["model"], temperature=role_setting["temperature"])


        agent = Runnable(agent_execution_mode='FuncCall', # 'FuncCall' or 'ReAct'，大模型支持FuncCall的话就用FuncCall
                                                tools=tools,
                                                llm=chat,
                                                assistant_message=role_setting["prompt"],)

        history += [HumanMessage(content=message)]
        result = asyncio.run(run_agent(agent, history))
        output("assistant", result, history)

